/********************************************************
* RANDOM TURING MACHINE SIMULATION                      *
* THEORY OF COMPUTER SCIENCE                            *
* PROF.: Joaquim Quinteiro Ucha [joukim@comp.ufla.br]  *
*                                                       *
* DEVELOPERS:                                           *
* Anderson de Rezende Rocha [undersun@comp.ufla.br]     *
* Jlio Csar Alves         [jcalves@comp.ufla.br]      *
*                                                       *
* BOTH computer scientists undergraduating students   at  *
* UFLA, Lavras Federal University at Lavras, MG, BRAZIL *
********************************************************/

#include "randomTuring.h"
#include <vector>
#include <fstream>

using namespace std;

// This file contains two classes:
// macro class defines a macro and its nodes to link in other macros
// rtMacros class from a file in the owner format mades one RTM in the 'machine.tmp' file.

// File format
// line 1: machine's description
// line 2 to ahead:
// p m n n2
// p = previous node
// m = macro's identifier
// n = next node
// n2 = next node2, only used by BRANCH macro, in this case, n identifies the output node when the
// number in the tape is greater than zero and n2 the output when the number in tape is zero

// The file allows line's comments identified by '!' character in the beginning of line.
// THE FILE DO NOT ALLOW BLANCK LINES.

class macro
{
    // The 0 (ZERO) node is always the FIRST, and all the numbers from 0 to the higher node identifier must be
    // used at least one time.
    public:
        macro(uint prevNode = 0, uint macroID = 0,uint nextNod = 0, uint nextNod2 = 0, uint lessStateID = 0)
        {
            setAll(prevNode, macroId, nextNod, nextNod2, lessStateID);
        }

        ~macro(){}
        // gets
        uint getPrevNode() { return previousNode; }
        uint getMacroId() { return macroId; }
        uint getNextNode() { return nextNode; }
        uint getNextNode2() { return nextNode2; }
        uint getLStateId() { return lessStateId; }

        //sets
        void setAll(uint prevNode = 0, uint macroID = 0,uint nextNod = 0, uint nextNod2 = 0, uint lessStateID = 0)
        {
            setPrevNode(prevNode);
            setMacroId(macroID);
            setNextNode(nextNod);
            setNextNode2(nextNod2);
            setLStateId(lessStateID);
        }

        void setPrevNode(uint prevNode) { previousNode = prevNode; }
        void setMacroId(uint macroID) { macroId = macroID; }
        void setNextNode(uint nextNod) { nextNode = nextNod; }
        void setNextNode2(uint nextNod2) { nextNode2 = nextNod2; }
        void setLStateId(uint lessStateID) { lessStateId = lessStateID; }

    private:
        uint previousNode; // identifies a macro's previous node
        uint macroId; // identifies a macro's identifier from database ('macros.dba' file)
        uint nextNode; // identifies a macro's next node
        uint nextNode2; // only used by BRANCH macro
        uint lessStateId; // Used by rtMacros class: the macro's q0 state will be q(0+lessStateId) in
                          // the final machine
};


// rtMacros class
template <class TOutput>
class rtMacros
{
    public:
        rtMacros();
        ~rtMacros();

        // RTM's functions

        int proccess( uint step = 0, bool print = false )
        {
            return machine->proccess(step, print);
        }

        int parser( vector<char> aInput, uint step = 0, bool print = false )
        {
            return machine->parser(aInput, step, print);
        }

        string getDescription()
        {
            return machine->getDescription();
        }

        void setDescription( string desc )
        {
            machine->setDescription(desc);
        }

        vector<char> getTape()
        {
            return machine->getTape();
        }

        void setTape( vector<char> i )
        {
            machine->setTape(i);
        }

        int getCursor()
        {
            return machine->getCursor();
        }

        uint getNStates()
        {
            return machine->getNStates();
        }

        flag getStatus()
        {
            return machine->getStatus();
        }

        uint getTLimit()
        {
            return machine->getTLimit();
        }

        void setTLimit( uint t )
        {
            machine->setTLimit(t);
        }

        uint getTCounter()
        {
            return machine->getTCounter;
        }

        bool readFile(string fileName);

        void print()
        {
            machine->print();
        }

    private:
        // creates a machine
        bool createMachine();

        // macros.dba
        // ! comments
        // macroId
        // macro description
        // transictions
        //string dataBaseName("macros.dba");

        // contains the macros' identifiers
        vector<macro> *macros;

        // the Random Turing Machine
        randomTuring<char, TOutput> *machine;
};

template <class TOutput>
rtMacros<TOutput>::rtMacros()
{
    vector<char> sigma(1);
    sigma[0] = '1'; // the alphabet
    vector<char> tau(1);
    tau[0] = '0'; // the auxiliary alphabet

    // alocating the RTM
    machine = new randomTuring<char, TOutput>;

    // setting alphabets
    machine->setSigma(sigma);
    machine->setTau(tau);

    macros = new vector<macro>(0);
}

template <class TOutput>
rtMacros<TOutput>::~rtMacros()
{
    delete machine;
    delete macros;
}


// make a machine from a file
template <class TOutput>
bool rtMacros<TOutput>::readFile(string fileName)
{
    macro aMacro;
    char machineDescription[50];
    uint prevNode, macroId, nextNode, nextNode2;

    // passing string to char...
    char charFileName[fileName.length()+1];
    fileName.copy(charFileName, fileName.length());
    charFileName[fileName.length()] = 0;
    // ... because the ifstream's constructor doesn't receive a string
    ifstream file(charFileName);

    if (!file)
    {
        cerr << "Error: File '" << fileName << "' not found" << endl;
        return false;
    }
    else // if file was opened
    {
        prevNode = macroId = nextNode = 0;
        while (file.peek() == '!') // ignoring comments
        {
            while (file.peek() != '\n')
                file.ignore();
            file.ignore();
        }

        // Reading ans setting the description
        file.getline(machineDescription, sizeof(machineDescription));
        setDescription(string (machineDescription));

        while (file.peek() == '!') // ignoring comments
        {
            while (file.peek() != '\n')
                file.ignore();
            file.ignore();
        }

        while (!file.eof())
        {
            nextNode2 = 0;

            // reading the identifiers
            file >> prevNode;
            file >> macroId;
            file >> nextNode;
            if (macroId == 0) // if it is the BRANCH macro
                file >> nextNode2; // reading the other node

            // setting the identifiers
            aMacro.setAll(prevNode, macroId, nextNode, nextNode2);

            // the vector contains the macro's identifiers by previous node,
            // thus a macro whose previous node is 'n' is in 'n' vector position
            if (macros->size() <= prevNode)
                macros->resize(prevNode+1);
            (*macros)[prevNode] = aMacro;

            while (file.peek() == '!') // ignoring comments
            {
                while (file.peek() != '\n')
                    file.ignore();
                file.ignore();
            }
        }
    }
    return createMachine();
}

// creates a machine from the macro's identifiers' vector and the database
template <class TOutput>
bool rtMacros<TOutput>::createMachine()
{
    int macroId;
    char line[100]; // to ignoring
    int finalStateId;
    int realFinalState = 0; // complete machine's final state
    int previousLessState;
    int aux; // used to various purposes
    char symbol;

    // this file contains the complete RTM made from the macros
    ofstream file("machine.tmp", ios::out);

    // this file contains all possible macros
    ifstream db("macros.dba");

    if (!file || !db)
    {
        if (!file)
            cerr << "Error: File 'machine.tmp' couldn't to be created" << endl;
        if (!db)
            cerr << "Error: Macros' Database ('macros.dba' file) couldn't to be opened" << endl;
        return false;
    }
    else // if files were opened
    {
        // setting lessStateIds
        previousLessState = 0;
        for(int i = 0; i < macros->size(); i++)
        {
            db.seekg(0); // going back to file's beginning
            db >> macroId; // reading the macro's identifier
            while (!db.eof() && macroId != (*macros)[i].getMacroId())
            {
                // if isn't the macro looked for, finding the next macro
                while (db.peek() != '?' && !db.eof())
                    db.getline(line, sizeof(line));
                db.getline(line, sizeof(line));
                db >> macroId;
            }
            if (db.eof())
            {
                cerr << "Macro's identifier '" << (*macros)[i].getMacroId() << "' doesn't exists" << endl;
                return false;
            }
            else
            {
                (*macros)[i].setLStateId(previousLessState);

                // ignoring lines
                db.getline(line, sizeof(line));
                db.getline(line, sizeof(line));

                db >> finalStateId; // reading the final state identifier
                previousLessState += finalStateId+1; // setting the previous less state

                while (db.peek() != '?' && !db.eof())
                    db.getline(line, sizeof(line));

                db.getline(line, sizeof(line));
            }
        }


        // generating the RTM file

        // writting the description
        file << getDescription() << endl;
        // writting the RTM's identifier (2), the alphabet(1) and the aux alphabet (0)
        file << "2\n1\n0\n";

        // calculating the complete machine's final state
        realFinalState = finalStateId+(*macros)[macros->size()-1].getLStateId()+1;
        file << realFinalState << endl;

        for(int i = 0; i < macros->size(); i++)
        {
            db.seekg(0);
            db >> macroId;
            // nao esta verificando se nao achou a macro
            while (macroId != (*macros)[i].getMacroId())
            {
                while (db.peek() != '?' && !db.eof())
                    db.getline(line, sizeof(line));
                db.getline(line, sizeof(line));
                db >> macroId;
            }
            db.getline(line, sizeof(line));
            db.getline(line, sizeof(line));
            // reading the transictions

            db >> finalStateId;
            db.getline(line, sizeof(line));

            // writing transictions
            while (!db.eof() && db.peek() != '?')
            {
                db >> aux;
                file << aux+(*macros)[i].getLStateId() << " ";

                db >> symbol; // reading the readSymbol
                file << symbol << " ";

                db >> aux; // reading the target state's id
                file << aux+(*macros)[i].getLStateId() << " ";

                db >> symbol;
                file << symbol << " "; // reading the write symbol

                db >> aux; // reading function's id
                file << aux << " ";

                db >> aux; // reading a register
                file << aux;

                // Verifying if there is any function's parameter
                if (!db.eof() && db.peek() != '\n')
                {
                    db >> aux; // reading a register
                    file << " " << aux;
                    // Verifying if there is more function's parameter
                    if (!db.eof() && db.peek() != '\n')
                    {
                        db >> aux; // regForMove
                        file << " " << aux;
                    }
                }
                file << endl;
                db.getline(line, sizeof(line));
            }

            if ((*macros)[i].getMacroId() == 0) // if it is Branch macro
            {
                file << finalStateId+(*macros)[i].getLStateId()-1 << " 0 ";
                //if ((*macros)[i].getNextNode() == macros->size())
                //    file << finalStateId+(*macros)[i].getLStateId()+1;
                //else
                file << (*macros)[(*macros)[i].getNextNode2()].getLStateId();
                file << " 0 0 0" << endl;
            }

            file << finalStateId+(*macros)[i].getLStateId() << " 0 ";
            if ((*macros)[i].getNextNode() == macros->size())
                file << realFinalState;
                //file << finalStateId+(*macros)[i].getLStateId()+1;
            else
                file << (*macros)[(*macros)[i].getNextNode()].getLStateId();
            file << " 0 0 0" << endl;

        }
        file.close();

        // the RTM file
        machine->readFile("machine.tmp");
    }
    return true;
}

